from PyPDF2 import PdfReader
from nltk.stem.snowball import SnowballStemmer
from nltk.tokenize import word_tokenize
import string
from operator import itemgetter
from mymodule import hinges, keywords
import re


class Text():
    """
    Prepares a text as keyword-dictionary

    Parameters: ".pdf or .txt filename", (first_page(int), last_page(int)), lang= "e"/"d"

    Class attributes:
    self.name
    self.language
    self.keywords
    self.hinges
    self.dict
    self.text

    """

    def __init__(self, filename, pages=False):
        # SET SELF
        self.name = filename
        self.language = "german"
        self.keywords = keywords
        self.hinges = hinges
        self.dict = {}

        # READ TXT-FILE INTO SELF.TEXT
        if filename.endswith('.txt'):
            with open(f"files/{filename}", "r") as raw:
                self.text = raw.read()

        # READ PDF-FILE INTO SELF.TEXT
        elif filename.endswith('.pdf'):  # .pdf file
            reader = PdfReader(f"files/{filename}")
            text = reader.pages
            self.text = ""
            if pages:
                text = text[pages[0]-1:pages[1]]
            for page in text:
                self.text += page.extract_text()
        self.text = (self.text.replace('\n', ' '))
        self.text = (self.text.replace('- ', ' '))

        # CREATE LIST OF SENTENCES FROM SELF.TEXT
        sent_separ = [".", "?", "!", ";", ":"]
        list_sent = []
        sentence = ""
        for char in self.text:
            if char not in sent_separ:
                sentence += char
            else:
                sentence += char
                list_sent.append(sentence)
                sentence = ""

        # CREATE DICTIONARY –> {KEYWORD1: {HINGE1: [SENTENCE1, SENTENCE2, …], HINGE2: [SENTENCE3, SENTENCE4, …]}, KEYWORD2: …}
        sst = SnowballStemmer(self.language)
        for sent in list_sent:
            for key in keywords:
                for hinge in hinges:
                    sent_token = word_tokenize(sent)
                    sent_token_stem = [sst.stem(w).lower() for w in sent_token]
                    hinge_stem = sst.stem(hinge).lower()
                    key_stem = sst.stem(key).lower()
                    if key_stem in sent_token_stem and hinge_stem in sent_token_stem:
                        if key not in self.dict:
                            self.dict[key] = {}
                            self.dict[key][hinge] = []
                            self.dict[key][hinge].append(sent)
                        else:
                            if hinge not in self.dict[key]:
                                self.dict[key][hinge] = []
                                self.dict[key][hinge].append(sent)
                            else:
                                self.dict[key][hinge].append(sent)


class Fusion(Text):
    """
    input:
    text1, text2 –> Text()

    attributes:
    self.dict
    self.text1
    self.text2
    self.name

    functions:
    display(): displays fusion on terminal
    write_file(): creates new .txt file with fusions
    """

    def __init__(self, text1, text2, length=None):
        self.length = length
        self.dict = self.__fusion(text1, text2, length)
        self.text1 = text1
        self.text2 = text2
        self.language = text1.language
        self.name = f"FUSION OF {text1.name.upper()} AND {text2.name.upper()}"

    def __fusion(self, text1, text2, length):
        """
        Fusion of two texts of Text() class.

        Parameters:
        arg1 (Text()), arg1 (Text())
        permutation=True/False: all combinations yes/no

        Returns:
        self.dict{"keyword": ["concat_sent1", "concat_sent2", …]}
        {KEYWORD1: {HINGE1: [concat_sent1, concat_sent2, …], HINGE2: [concat_sent3, concat_sent4, …]}, KEYWORD2: …}
        """

        def rebuild(i_range, sent_token):  # rebuild half sentences
            half_sent = " "

            for n in i_range:
                word = sent_token[n]
                half_sent = half_sent + " " + word
            return half_sent

        def cut(sent, side):  # CUT AT KEYWORD
            sst = SnowballStemmer("german")

            # prepare sentence, get keyword
            sent_token = word_tokenize(sent)
            sent_token_stem = [sst.stem(w).lower() for w in sent_token]
            hinge_stem = sst.stem(hinge)
            i = sent_token_stem.index(hinge_stem)  # keyword index
            # check if hinge at beginning or end of sentence
            if i == 0 or i == len(sent_token_stem)-2:
                result = "ignore"
            else:
                if side == "left":  # work left side
                    i_range = range(i+1)
                    result = rebuild(i_range, sent_token)
                if side == "right":  # work right side
                    i_range = range(i+1, len(sent_token))
                    right = rebuild(i_range, sent_token)
                    result = right[1:]
            return result

        def unite_two(sent1, sent2):  # 1 LEFT SIDE + 1 RIGHT SIDE
            # CONCAT HALFS
            raw_sent = cut(sent1, side="left") + cut(sent2, side="right")
            # REMOVE DOUBLE SPACES
            return re.sub(' {2,}', ' ', raw_sent)

        dict_fusions = {}
        # ALL COMBINATIONS OF HALF SENTENCES

        common_keywords = text1.dict.keys() & text2.dict.keys()
        for keyword in common_keywords:
            dict_fusions[keyword] = {}  # create list per key
            common_hinges = text1.dict[keyword].keys(
            ) & text2.dict[keyword].keys()
            for hinge in common_hinges:
                dict_fusions[keyword][hinge] = []  # create list per hinge

                for s_left in text1.dict[keyword][hinge]:  # iteration in list1
                    # iteration in list2
                    for s_right in text2.dict[keyword][hinge]:
                        fusion_1 = unite_two(s_left, s_right).strip()  # regual
                        fusion_2 = unite_two(
                            s_right, s_left).strip()  # reversed
                        if length:  # case length is defined
                            if len(fusion_1) <= length and "ignore" not in fusion_1:
                                dict_fusions[keyword][hinge].append(fusion_1)
                            if len(fusion_2) <= length and "ignore" not in fusion_2:
                                dict_fusions[keyword][hinge].append(fusion_2)
                        else:  # case no length restictions
                            if "ignore" not in fusion_1:
                                dict_fusions[keyword][hinge].append(fusion_1)
                            if "ignore" not in fusion_2:
                                dict_fusions[keyword][hinge].append(fusion_2)
        return dict_fusions

    def __format_fusion(self):
        # HEADER
        names = (self.text1.name + " + " + self.text2.name).upper()
        line = (len(names)+3)*"–"
        result = f"{line}\n{names}\n{line}\n\n"
        # KEYWORD HEADER + SENTENCES
        for keyword in self.dict:
            for hinge in self.dict[keyword]:
                name_nmbr = f"{keyword}/{hinge}: {str(len(self.dict[keyword][hinge]))}"
                line = (len(name_nmbr)+4)*"-"
                result += f"{line}\n{name_nmbr}\n{line}\n"
                for sent in self.dict[keyword][hinge]:
                    result = result + "• " + sent + "\n\n"
        return result

    def display(self):
        """
        Prints info + sentences.
        """
        output = self.__format_fusion()
        print(output)

    def write_file(self, filename):
        """
        writes file as .txt:
        info + nmbr sentences
        """
        output = self.__format_fusion()
        with open(f"results/{filename}.txt", "w") as file:
            file.write(output)
